#include "cmd.h"
#include "hardw.h"
#include "vars.h"

#define SCK F_11                      /* Serial clock */

/* -------90S1200P (program memory) serial downloading routines-------- */

char spi_in_out(char c)
{
  char cnt, b;

  b = 0;
  for (cnt = 0; cnt < 8; cnt++)
  {
    if (c & 0x80) SetF(F_16, 1);
    else SetF(F_16, 0);               /* Send bits to MOSI starting from MSB */
    c <<= 1;                          /* Shift bits to the left */
    /* delay10us(2); */                     /* Wait for 20 us */
    SCK = 1;                          /* Give a pulse to serial clock input */
    /* delay10us(2); */                     /* Wait for 20 us */
    b <<= 1;                          /* Shift bits to the left */
    if (F_4) b |= 1;                  /* Receive bits from MISO */
    SCK = 0;
    /* delay10us(2); */                     /* Wait for 20 us */
  }
  return(b);
}

char ch90S1200P_init(char c)
{
  init_setting();                     /* Call common init procedure */
  SCK = 1;                            /* Clock = High */
  PPA1[KA] = 2;                       /* Give RESET a positive pulse */
  SetF(F_16, 1);
  delay1ms(20);                       /* Wait for 20 ms */
  SCK = 0;                            /* Clock = Low */
  delay1ms(10);                       /* Wait for 10 ms */
  PPA1[KA] = 0;
  delay1ms(40);                       /* Wait for 40 ms */
  spi_in_out(0xAC);                   /* 10101100 */
  spi_in_out(0x53);                   /* 01010011 */
  spi_in_out(0);                      /* Programming enable */
  spi_in_out(0);                      /* serial instruction */
}

char ch90S1200P_set_addr(char c)
{
  return(1);
}

char ch90S1200P_read(char c)
{
  if (msg.cmd.address.lc.LO_LO_CHAR & 1)       /* High (odd) byte */
    c = 0x28;                                  /* 00101000 */
  else c = 0x20;                               /* 00100000 */
  spi_in_out(c);                               /* Send Read instruction */
  cnt = msg.cmd.address.lw.LO_WORD >> 1;       /* Get current word address */
  spi_in_out(hi_byte(cnt));                    /* Send address high byte */
  spi_in_out(lo_byte(cnt));                    /* Send address low byte */
  return(spi_in_out(0));                       /* Read data from MISO */
}

char ch90S1200P_write(char c)
{
  if (msg.cmd.address.lc.LO_LO_CHAR & 1)       /* High (odd) byte */
    spi_in_out(0x48);                          /* 01001000 */
  else spi_in_out(0x40);                       /* 01000000 */
  cnt = msg.cmd.address.lw.LO_WORD >> 1;       /* Get current byte address */
  spi_in_out(hi_byte(cnt));                    /* Send address high byte */
  spi_in_out(lo_byte(cnt));                    /* Send address low byte */
  spi_in_out(c);                               /* Send data to MOSI */
  delay1ms(4);                                 /* Wait for 4 ms */
  return(1);
}

char ch90S1200P_erase(char c)
{
  if (c < eLockEx)
  {
    ch90S1200P_init(1);
    spi_in_out(0xAC);                          /* 10101100 */
    if ((c == eErase) || (c == eUErase))
      spi_in_out(0x80);                        /* 10000000 Erase instruction */
    else if (c == eLockRead)
      spi_in_out(0xE0);                        /* 11100000 Lock bits */
    else if (c == eLockWrite)
      spi_in_out(0xE4);                        /* 11100100 Lock bits */
    spi_in_out(0);                             /* Send something */
    spi_in_out(0);
    delay1ms(10);                              /* Wait for 10 ms */
  }
  return(1);
}

/* -------90S1200E (EEPROM) serial downloading routines-------- */

char ch90S1200E_read(char c)
{
  spi_in_out(0xA0);                            /* 10100000 */
  spi_in_out(0);                               /* Send Read instruction */
  spi_in_out(msg.cmd.address.lc.LO_LO_CHAR);   /* Send address byte */
  return(spi_in_out(0));                       /* Read data from MISO */
}

char ch90S1200E_write(char c)
{
  spi_in_out(0xC0);                            /* 10100000 */
  spi_in_out(0);                               /* Send Write instruction */
  spi_in_out(msg.cmd.address.lc.LO_LO_CHAR);   /* Send address byte */
  spi_in_out(c);                               /* Send data to MOSI */
  delay1ms(4);                                 /* Wait for 4 ms */
  return(1);
}