{$F+,O+,I-,X+}
unit ApPrim;

interface

uses APCom, APPort;

const
  ComName: ComNameType = Com2;
  Baud: Longint = 9600;
  Parity: ParityType = NoParity;
  DataBits: DataBitType = 8;
  StopBits: StopBitType = 2;
  HWFlow: Word = 0;
  InBufSize: Word = 4096;
  OutBufSize: Word = 4096;
  PortOptions: Word = DefPortOptionsSimple;
  TryTimeout: Word = 18;
  FullTimeout: Word = $FFF0;
  LastModemStatus: Byte = 0;
  CancelReq: Boolean = False;
  CancelNow: Boolean = False;

  RollerTicks: Word = 0;
  RollerCount: Byte = 0;
  RollerIndex: Byte = 0;
  RollerChars: Array[0..3] of Char = ('-', '\', '|', '/');
  RollerBack:  Boolean = False;
  RollerInUse: Boolean = False;

procedure InitRoller;
procedure DoneRoller;

function KbdAbort: Boolean;
procedure AsyncErrorProc(P: Pointer; var Code: Word);

procedure InitPortPrim(var P: PortRecPtr; ComName: ComNameType; Baud: LongInt;
  Parity: ParityType; DataBits: DataBitType; StopBits: StopBitType;
  InSize, OutSize: Word; Options: Word);
procedure DonePortPrim(var P: PortRecPtr);
function GetCharPrim(P: PortRecPtr; var C: Char): Boolean;
function PutCharPrim(P: PortRecPtr; C: Char): Boolean;

implementation

uses Crt, Win, APMisc, DevLayer;

procedure InitRoller;
begin
  RollerTicks := MemW[$40:$6C];
  RollerCount := 0;
  RollerIndex := 0;
  RollerBack  := False;
  RollerInUse := True;
end;

procedure WriteRoller(C: Char);
begin
  WriteStr(2, 3, C, $7A);
  WriteStr(43, 3, C, $7A);
end;

procedure DoneRoller;
begin
  WriteRoller(' ');
  RollerInUse := False;
end;

procedure UpdateRoller;
var
  W: Word;
begin
  if not RollerInUse then Exit;
  if RollerCount < 8 then Inc(RollerCount)
  else
  begin
    W := MemW[$40:$6C] - RollerTicks;
    if (RollerCount = 8) and (W > 32) then DoneRoller
    else
    begin
      RollerCount := $FF;
      if W >= 2 then
      begin
	RollerTicks := MemW[$40:$6C];
	WriteRoller(RollerChars[RollerIndex]);
	if RollerBack then
	  if RollerIndex = 0 then RollerIndex := 3
	  else Dec(RollerIndex)
	else
	  if RollerIndex = 3 then RollerIndex := 0
	  else Inc(RollerIndex);
      end;
    end;
  end;
end;

function KbdAbort: Boolean;
var
  B: Boolean;
begin
  UpdateRoller;
  B := CancelReq;
  if KeyPressed and (ReadKey = #27) then
  begin
    CancelReq := True;
    KbdAbort := CancelNow or (B and CancelReq);
  end
  else KbdAbort := False;
end;

procedure AsyncErrorProc(P: Pointer; var Code: Word);
begin
end;

procedure InitPortPrim(var P: PortRecPtr; ComName: ComNameType; Baud: LongInt;
  Parity: ParityType; DataBits: DataBitType; StopBits: StopBitType;
  InSize, OutSize: Word; Options: Word);
var
  B: Boolean;
  I: Integer;
begin
  DeviceLayer := dlUART;
  InitDeviceLayer;
  B := False;
  for I := 1 to MaxActivePort do
    if (ActiveComPort[I] <> nil) and
       (ActiveComPort[I]^.PortName = ComName) then
    begin
      B := True;
      Break;
    end;
  if B then AsyncStatus := ecNoMorePorts
  else InitPort(P, ComName, Baud, Parity, DataBits, StopBits,
    InSize, OutSize, Options);
  if AsyncStatus <> ecOk then AsyncErrorProc(P, AsyncStatus)
  else
  begin
    SWFlowDisable(P);
    if HWFlow = 0 then HWFlowDisable(P)
    else HWFlowEnable(P, Round(InSize * 0.75), Round(InSize * 0.25), HWFlow);
    SetAbortFunc(P, KbdAbort);
    SetErrorProc(P, AsyncErrorProc);
  end;
end;

procedure DonePortPrim(var P: PortRecPtr);
begin
  DonePort(P);
end;

function GetCharPrim(P: PortRecPtr; var C: Char): Boolean;
begin
  GetCharPrim := False;
  if not CharReady(P) then Exit;
  GetChar(P, C);
  GetCharPrim := AsyncStatus = ecOk;
end;

function PutCharPrim(P: PortRecPtr; C: Char): Boolean;
var
  S: String;
begin
  PutCharPrim := False;
  APCom.PutCharTimeout(P, C, TryTimeout);
  if AsyncStatus <> ecOk then
    APCom.PutCharTimeout(P, C, FullTimeout);
  PutCharPrim := AsyncStatus = ecOk;
end;

end.
